#!/usr/bin/env python3
# SPDX-FileCopyrightText: 2022 The Ikarus Developers mueller@ibb.uni-stuttgart.de
# SPDX-License-Identifier: CC0-1.0
import os
from datetime import datetime


def read_old_version():
    script_dir = os.path.dirname(os.path.realpath("__file__"))
    rel_path = "dune.module"
    abs_file_path = os.path.join(script_dir, rel_path)
    print(abs_file_path)
    with open(abs_file_path) as f:
        s = f.read()
        for line in s.split("\n"):
            print(line)
            if line.startswith("Version: "):
                _, old_version = line.split(" ")
                return old_version


def bump_patch_number(version_number: str) -> str:
    """Return a copy of `version_number` with the patch number incremented."""
    major, minor, patch = version_number.split(".")
    return f"{major}.{minor}.{int(patch) + 1}"


def inplace_change(filename: str, old_string: str, new_string: str):
    with open(filename) as f:
        s = f.read()
        if old_string not in s:
            print('"{old_string}" not found in {filename}.'.format(**locals()))
            return

    with open(filename, "w") as f:
        print(
            'Changing "{old_string}" to "{new_string}" in {filename}'.format(**locals())
        )
        s = s.replace(old_string, new_string)
        f.write(s)


def changeLine(filename: str, old_string: str, new_string: str):
    with open(filename) as f:
        s = f.read()
        for line in s.split("\n"):
            print(line)
            if line.startswith(old_string):
                with open(filename, "w") as fw:
                    print(
                        'Changing the line "{line}"\n to "{new_string}" in {filename}'.format(
                            **locals()
                        )
                    )
                    s = s.replace(line, new_string)
                    fw.write(s)


def update_all_versions(version_override=None):
    """Update all version numbers in local files"""
    old_version_number = read_old_version()
    if version_override == "dev":
        new_version_number = bump_patch_number(old_version_number)
    elif version_override is None:
        new_version_number = old_version_number
    else:
        new_version_number = version_override

    if version_override == "dev":
        new_version_number += ".dev" + datetime.now().strftime("%Y%m%d%H%M%S")

    print(f"Bump version from {old_version_number} to {new_version_number}")
    if version_override != "dev":
        inplace_change(
            "dune.module",
            f"Version: {old_version_number}",
            f"Version: {new_version_number}",
        )
        inplace_change(
            "CMakeLists.txt",
            f"VERSION {old_version_number}",
            f"VERSION {new_version_number}",
        )

    changeLine(
        "setup.py",
        f"ikarusVersion =",
        f'ikarusVersion = "{new_version_number}"',
    )


import sys

if __name__ == "__main__":
    try:
        var = sys.argv[1]
        var = var.removeprefix("v")
        update_all_versions(var)
    except IndexError:
        update_all_versions()



# SPDX-FileCopyrightText: 2022 The Ikarus Developers mueller@ibb.uni-stuttgart.de
# SPDX-License-Identifier: LGPL-3.0-or-later

import setpath

setpath.set_path()
import ikarus as iks
import ikarus.finite_elements
import ikarus.utils
import ikarus.assembler
import ikarus.dirichlet_values
import numpy as np
import scipy as sp
from scipy.optimize import minimize

import dune.grid
import dune.functions

if __name__ == "__main__":
    lowerLeft = []
    upperRight = []
    elements = []
    for i in range(2):
        lowerLeft.append(-1)
        upperRight.append(1)
        elements.append(3)

    req = ikarus.FErequirements()
    req.addAffordance(iks.ScalarAffordances.mechanicalPotentialEnergy)

    grid = dune.grid.structuredGrid(lowerLeft, upperRight, elements)
    grid.hierarchicalGrid.globalRefine(0)
    basisLagrange1 = ikarus.basis(
        grid, dune.functions.Power(dune.functions.Lagrange(order=1), 2)
    )
    flatBasis = basisLagrange1.flat()

    forces = np.zeros(8)
    stiffness = np.zeros((8, 8))

    def volumeLoad(x, lambdaVal):
        return np.array([lambdaVal * x[0] * 2, 2 * lambdaVal * x[1] * 0])

    def neumannLoad(x, lambdaVal):
        return np.array([lambdaVal * 0, lambdaVal])

    neumannVertices = np.zeros(grid.size(2) * 2, dtype=bool)
    assert len(neumannVertices) == len(flatBasis)

    flatBasis.interpolate(neumannVertices, lambda x: True if x[1] > 0.9 else False)

    boundaryPatch = iks.utils.boundaryPatch(grid, neumannVertices)

    svk = iks.materials.StVenantKirchhoff(E=1000, nu=0.3)

    psNH = svk.asPlaneStress()
    fes = []
    for e in grid.elements:
        fes.append(
            iks.finite_elements.NonLinearElastic(
                basisLagrange1, e, psNH, volumeLoad, boundaryPatch, neumannLoad
            )
        )

    dirichletValues = iks.dirichletValues(flatBasis)

    def fixFirstIndex(vec, globalIndex):
        vec[0] = True

    def fixAnotherVertex(vec, localIndex, localView):
        localView.index(localIndex)
        vec[1] = True

    def fixLeftHandEdge(vec, localIndex, localView, intersection):
        if intersection.geometry.center[1] < -0.9:
            vec[localView.index(localIndex)] = True

    dirichletValues.fixBoundaryDOFs(fixFirstIndex)
    dirichletValues.fixBoundaryDOFsUsingLocalView(fixAnotherVertex)
    dirichletValues.fixBoundaryDOFsUsingLocalViewAndIntersection(fixLeftHandEdge)

    assembler = iks.assembler.sparseFlatAssembler(fes, dirichletValues)

    dRed = np.zeros(assembler.reducedSize())

    lambdaLoad = iks.ValueWrapper(3.0)

    def energy(dRedInput):
        reqL = ikarus.FErequirements()
        reqL.addAffordance(iks.ScalarAffordances.mechanicalPotentialEnergy)
        reqL.insertParameter(iks.FEParameter.loadfactor, lambdaLoad)

        dBig = assembler.createFullVector(dRedInput)
        reqL.insertGlobalSolution(iks.FESolutions.displacement, dBig)
        return assembler.getScalar(reqL)

    def gradient(dRedInput):
        reqL = ikarus.FErequirements()
        reqL.addAffordance(iks.VectorAffordances.forces)
        reqL.insertParameter(iks.FEParameter.loadfactor, lambdaLoad)

        dBig = assembler.createFullVector(dRedInput)
        reqL.insertGlobalSolution(iks.FESolutions.displacement, dBig)
        return assembler.getReducedVector(reqL)

    def hess(dRedInput):
        reqL = ikarus.FErequirements()
        reqL.addAffordance(iks.MatrixAffordances.stiffness)
        reqL.insertParameter(iks.FEParameter.loadfactor, lambdaLoad)

        dBig = assembler.createFullVector(dRedInput)
        reqL.insertGlobalSolution(iks.FESolutions.displacement, dBig)
        return assembler.getReducedMatrix(reqL).todense()

    resultd = minimize(energy, x0=dRed, options={"disp": True}, tol=1e-14)
    resultd2 = minimize(
        energy, x0=dRed, jac=gradient, options={"disp": True}, tol=1e-14
    )
    resultd3 = minimize(
        energy,
        method="trust-constr",
        x0=dRed,
        jac=gradient,
        hess=hess,
        options={"disp": True},
    )
    resultd4 = sp.optimize.root(gradient, jac=hess, x0=dRed, tol=1e-10)

    np.set_printoptions(precision=3)

    assert np.allclose(resultd.x, resultd2.x, atol=1e-6)
    assert np.allclose(resultd3.x, resultd4.x)
    assert np.all(abs(resultd3.grad) < 1e-8)
    assert np.all(abs(resultd4.fun) < 1e-8)

    resReq = ikarus.ResultRequirements()
    resReq.insertGlobalSolution(
        iks.FESolutions.displacement, assembler.createFullVector(resultd.x)
    )
    resReq.insertParameter(iks.FEParameter.loadfactor, lambdaLoad)
    resReq.addResultRequest(iks.ResultType.PK2Stress)

    res1 = fes[0].resultAt(resReq, np.array([0.5, 0.5]))
    res2 = fes[0].resultAt(resReq, np.array([0.5, 0.5]), iks.ResultType.PK2Stress)

    assert np.allclose(res1, res2)



# SPDX-FileCopyrightText: 2022 The Ikarus Developers mueller@ibb.uni-stuttgart.de
# SPDX-License-Identifier: LGPL-3.0-or-later

from pathlib import Path

OUTPUT_PATH = "ikarus/python/test/"


# @todo define a common place where python tests should output the output files to
def output_path():
    cwd = Path.cwd()

    if cwd.name == "test":
        return ""
    else:
        return OUTPUT_PATH



# SPDX-FileCopyrightText: 2022 The Ikarus Developers mueller@ibb.uni-stuttgart.de
# SPDX-License-Identifier: LGPL-3.0-or-later

from dune.generator.generator import SimpleGenerator


def decoratePre(pre):
    def wrappedPre(*args, **kwargs):
        preamble = pre(*args, **kwargs)
        newPreamble = ""
        newPreamble += "#define DUNE_LOCALFEFUNCTIONS_USE_EIGEN 1\n"
        newPreamble += "#define EIGEN_DEFAULT_TO_ROW_MAJOR 1\n"  # needed to have conforming Matrix storage between eigen and numpy otherwise references are not working
        newPreamble += preamble
        return newPreamble

    return wrappedPre


myAttributes = vars(SimpleGenerator).copy()
myAttributes["pre"] = decoratePre(myAttributes["pre"])
MySimpleGenerator = type("MySimpleGenerator", (object,), myAttributes)
